# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from concurrent import futures
import threading
import unittest

from lxm3._vendor.xmanager import xm_mock
from lxm3._vendor.xmanager.xm import core
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3._vendor.xmanager.xm import utils


class TestError(RuntimeError):
  """Exception which can be used in tests below."""


async def failing_job_generator(work_unit: core.WorkUnit):
  raise TestError


class ApplyArgsTest(unittest.TestCase):

  def test_wrong_job_args(self):
    with self.assertRaises(ValueError):
      core._apply_args(
          job_blocks.Job(
              job_blocks.Executable(name=''), xm_mock.MockExecutor()
          ),
          {'abra': 'kadabra'},
      )

  def test_wrong_job_group_args(self):
    with self.assertRaises(ValueError):
      core._apply_args(
          job_blocks.JobGroup(
              learner=job_blocks.Job(
                  job_blocks.Executable(name=''), xm_mock.MockExecutor()
              )
          ),
          {'eval': {'args': {'batch_size': 32}}},
      )


class ExperimentTest(unittest.TestCase):

  def test_single_job_launch(self):
    experiment = xm_mock.MockExperiment()
    with experiment:
      job = job_blocks.Job(
          xm_mock.MockExecutable(), xm_mock.MockExecutor(), args={}, name='name'
      )
      experiment.add(job)

    self.assertEqual(experiment.launched_jobs, [job])

  def test_job_group_launch(self):
    experiment = xm_mock.MockExperiment()
    with experiment:
      foo_job = job_blocks.Job(
          xm_mock.MockExecutable(),
          xm_mock.MockExecutor(),
          args={'foo': 1},
          name='1',
      )
      bar_job = job_blocks.Job(
          xm_mock.MockExecutable(),
          xm_mock.MockExecutor(),
          args={'bar': 2},
          name='2',
      )
      experiment.add(job_blocks.JobGroup(foo=foo_job, bar=bar_job))

    self.assertEqual(experiment.launched_jobs, [foo_job, bar_job])

  def test_job_generator_launch(self):
    experiment = xm_mock.MockExperiment()
    with experiment:
      job = job_blocks.Job(
          xm_mock.MockExecutable(), xm_mock.MockExecutor(), args={}, name='name'
      )

      async def job_generator(work_unit: core.WorkUnit, use_magic: bool):
        self.assertEqual(use_magic, True)
        work_unit.add(job)

      experiment.add(job_generator, args={'use_magic': True})

    self.assertEqual(experiment.launched_jobs, [job])
    self.assertEqual(experiment.launched_jobs_args, [{'use_magic': True}])

  def test_job_generator_raises(self):
    experiment = xm_mock.MockExperiment()
    with self.assertRaises(TestError):
      with experiment:
        experiment.add(failing_job_generator)

  def test_non_async_job_generator_raises_user_friendly_exception(self):
    with self.assertRaisesRegex(ValueError, '.* generator must be an async .*'):
      with xm_mock.MockExperiment() as experiment:

        def job_generator(work_unit: core.WorkUnit):
          del work_unit

        experiment.add(job_generator)

  def test_auxiliary_unit_job(self):
    experiment = xm_mock.MockExperiment()
    with experiment:
      job = job_blocks.Job(
          xm_mock.MockExecutable(), xm_mock.MockExecutor(), args={}, name='name'
      )
      experiment.add(core.AuxiliaryUnitJob(job, termination_delay_secs=600))

    self.assertEqual(len(experiment.auxiliary_units), 1)

  def test_auxiliary_unit_job_generator(self):
    experiment = xm_mock.MockExperiment()
    with experiment:

      async def make_job(aux_unit: core.ExperimentUnit):
        aux_unit.add(
            job_blocks.Job(
                xm_mock.MockExecutable(),
                xm_mock.MockExecutor(),
                args={},
                name='name',
            )
        )

      experiment.add(
          core.AuxiliaryUnitJob(make_job, termination_delay_secs=600)
      )

    self.assertEqual(len(experiment.auxiliary_units), 1)

  def test_launch_with_args(self):
    experiment = xm_mock.MockExperiment()
    with experiment:
      experiment.add(
          job_blocks.JobGroup(
              foo=job_blocks.Job(
                  xm_mock.MockExecutable(),
                  xm_mock.MockExecutor(),
                  args={'x': 1, 'y': 2},
                  env_vars={'EDITOR': 'vi'},
              ),
              bar=job_blocks.Job(
                  xm_mock.MockExecutable(),
                  xm_mock.MockExecutor(),
                  args=['--bar=1'],
              ),
          ),
          args={
              'foo': {'args': {'x': 3, 'z': 4}, 'env_vars': {'TURBO': 'ON'}},
              'bar': {'args': ['--spacebar']},
          },
      )

    self.assertEqual(
        experiment.launched_jobs[0].args,
        job_blocks.SequentialArgs.from_collection({'x': 3, 'y': 2, 'z': 4}),
    )
    self.assertEqual(
        experiment.launched_jobs[0].env_vars, {'TURBO': 'ON', 'EDITOR': 'vi'}
    )
    self.assertEqual(
        experiment.launched_jobs[1].args,
        job_blocks.SequentialArgs.from_collection(['--bar=1', '--spacebar']),
    )

  def test_launch_with_different_args(self):
    experiment = xm_mock.MockExperiment()
    with experiment:
      job = job_blocks.Job(xm_mock.MockExecutable(), xm_mock.MockExecutor())
      for i in range(10):
        experiment.add(job, args={'env_vars': {'FOO': i}})

    self.assertEqual(experiment.launched_jobs[0].env_vars, {'FOO': 0})
    self.assertEqual(experiment.launched_jobs[1].env_vars, {'FOO': 1})
    self.assertEqual(experiment.launched_jobs[2].env_vars, {'FOO': 2})

  def test_add_runs_asynchronously(self):
    generator_called = threading.Event()

    with xm_mock.MockExperiment() as experiment:

      async def job_generator(work_unit: core.WorkUnit):
        del work_unit
        generator_called.set()

      experiment.add(job_generator)

      # Validate that job_generator is executed in a parallel thread.
      self.assertTrue(generator_called.wait(timeout=5))

  @utils.run_in_asyncio_loop
  async def test_loop_is_reused_in_coro_context(self):
    loop = asyncio.get_event_loop()

    async with xm_mock.MockExperiment() as experiment:

      async def job_generator(work_unit: core.WorkUnit):
        del work_unit
        self.assertEqual(asyncio.get_event_loop(), loop)

      experiment.add(job_generator)

  @utils.run_in_asyncio_loop
  async def test_sync_with_cant_be_used_in_coro_context(self):
    # `async with` works.
    async with xm_mock.MockExperiment():
      pass

    with self.assertRaises(RuntimeError):
      # But `with` raises an exception.
      with xm_mock.MockExperiment():
        pass

  def test_experiment_works_from_thread_pool(self):
    # There would be no Asyncio even loop thread attahched if running from a
    # worker thread. We ensure that the API still works.
    def launch_experiment():
      experiment = xm_mock.MockExperiment()
      with experiment:
        experiment.add(
            job_blocks.Job(
                xm_mock.MockExecutable(), xm_mock.MockExecutor(), args={}
            )
        )

    with futures.ThreadPoolExecutor() as executor:
      executor.submit(launch_experiment).result()

  @utils.run_in_asyncio_loop
  async def test_work_unit_wait_until_complete(self):
    experiment = xm_mock.MockExperiment()
    async with experiment:
      experiment.add(
          job_blocks.Job(
              xm_mock.MockExecutable(), xm_mock.MockExecutor(), args={}
          )
      )
      completion_future = experiment.work_units[0].wait_until_complete()
      self.assertEqual(completion_future.work_unit.work_unit_id, 1)
      await completion_future

  @utils.run_in_asyncio_loop
  async def test_work_unit_wait_until_complete_exception(self):
    experiment = xm_mock.MockExperiment()
    with self.assertRaises(TestError):
      async with experiment:
        experiment.add(failing_job_generator)
        with self.assertRaises(core.ExperimentUnitError):
          await experiment.work_units[0].wait_until_complete()

  @utils.run_in_asyncio_loop
  async def test_get_full_job_name(self):
    async def generator(work_unit):
      self.assertEqual(work_unit.get_full_job_name('name'), '1_1_name')

    async with xm_mock.MockExperiment() as experiment:
      experiment.add(generator)


class ContextvarsTest(unittest.TestCase):

  def test_contextvars_single_job_launch(self):
    with xm_mock.MockExperiment() as experiment:
      job = job_blocks.Job(xm_mock.MockExecutable(), xm_mock.MockExecutor())
      self.assertEqual(core._current_experiment.get(), experiment)
      experiment.add(job)

    self.assertIsNone(core._current_experiment.get(None))
    self.assertIsNone(core._current_experiment_unit.get(None))

  def test_contextvars_job_group_launch(self):
    with xm_mock.MockExperiment() as experiment:
      foo_job = job_blocks.Job(xm_mock.MockExecutable(), xm_mock.MockExecutor())
      self.assertEqual(core._current_experiment.get(), experiment)
      experiment.add(job_blocks.JobGroup(foo=foo_job))

    self.assertIsNone(core._current_experiment.get(None))
    self.assertIsNone(core._current_experiment_unit.get(None))

  def test_contextvars_job_generator_launch(self):
    with xm_mock.MockExperiment() as experiment:
      self.assertEqual(core._current_experiment.get(), experiment)

      async def job_generator(work_unit: core.WorkUnit):
        self.assertEqual(core._current_experiment_unit.get(), work_unit)
        self.assertEqual(core._current_experiment.get(), work_unit.experiment)

      experiment.add(job_generator)

    self.assertIsNone(core._current_experiment.get(None))
    self.assertIsNone(core._current_experiment_unit.get(None))

  def test_contextvars_async_job_generator_launch(self):
    async def make_experiment():
      async with xm_mock.MockExperiment() as experiment:

        async def job_generator(work_unit: core.WorkUnit):
          self.assertEqual(core._current_experiment_unit.get(), work_unit)
          self.assertEqual(core._current_experiment.get(), work_unit.experiment)

        experiment.add(job_generator)
        self.assertEqual(core._current_experiment.get(), experiment)

      asyncio.run(make_experiment())

    self.assertIsNone(core._current_experiment.get(None))
    self.assertIsNone(core._current_experiment_unit.get(None))

  def test_contextvars_nested_async_job_generator_launch(self):
    async def job_generator(work_unit: core.WorkUnit):
      self.assertEqual(core._current_experiment.get(), work_unit.experiment)
      self.assertEqual(core._current_experiment_unit.get(), work_unit)

    with xm_mock.MockExperiment() as outer_exp:

      async def make_inner_exp():
        async with xm_mock.MockExperiment() as experiment:
          experiment.add(job_generator)
          outer_exp.add(job_generator)
          self.assertEqual(core._current_experiment.get(), experiment)

      asyncio.run(make_inner_exp())
      self.assertEqual(core._current_experiment.get(), outer_exp)

    self.assertIsNone(core._current_experiment.get(None))
    self.assertIsNone(core._current_experiment_unit.get(None))


if __name__ == '__main__':
  unittest.main()



import os
import subprocess
from typing import List, Optional, Sequence


def rsync(
    src: str,
    dst: str,
    opt: List[str],
    host: Optional[str] = None,
    excludes: Optional[Sequence[str]] = None,
    filters: Optional[Sequence[str]] = None,
    mkdirs: bool = False,
):
    if excludes is None:
        excludes = []
    if filters is None:
        filters = []
    opt = list(opt)
    for exclude in excludes:
        opt.append(f"--exclude={exclude}")
    for filter in filters:
        opt.append(f"--filter=:- {filter}")
    if not host:
        os.makedirs(os.path.dirname(dst), exist_ok=True)
        sync_cmd = ["rsync"] + opt + [src, dst]
        subprocess.check_call(sync_cmd)
    else:
        if mkdirs:
            subprocess.check_output(["ssh", host, "mkdir", "-p", os.path.dirname(dst)])
        dst = f"{host}:{dst}"
        sync_cmd = ["rsync"] + opt + [src, dst]
        subprocess.check_call(sync_cmd)



import datetime
import functools
import os
from typing import List, Optional

from lxm3 import xm
from lxm3.clusters import slurm
from lxm3.xm_cluster import array_job
from lxm3.xm_cluster import artifacts
from lxm3.xm_cluster import config as config_lib
from lxm3.xm_cluster import executables
from lxm3.xm_cluster import executors
from lxm3.xm_cluster.console import console
from lxm3.xm_cluster.execution import job_script


class SlurmJobScriptBuilder(job_script.JobScriptBuilder):
    TASK_OFFSET = 1
    JOB_SCRIPT_SHEBANG = "#!/usr/bin/bash -l"
    TASK_ID_VAR_NAME = "SLURM_ARRAY_TASK_ID"

    @classmethod
    def _is_gpu_requested(cls, executor: executors.Slurm) -> bool:
        del executor
        return True  # TODO

    @classmethod
    def _create_setup_cmds(cls, executable, executor) -> str:
        cmds = ["echo >&2 INFO[$(basename $0)]: Running on host $(hostname)"]

        for module in executor.modules:
            cmds.append(f"module load {module}")
        if cls._is_gpu_requested(executor):
            cmds.append(
                "echo >&2 INFO[$(basename $0)]: CUDA_VISIBLE_DEVICES=$CUDA_VISIBLE_DEVICES"
            )
            # cmds.append("nvidia-smi")

        if executable.singularity_image is not None:
            cmds.append(
                "echo >&2 INFO[$(basename $0)]: Singularity version: $(singularity --version)"
            )
        return "\n".join(cmds)

    @classmethod
    def _create_job_header(
        cls,
        executor: executors.Slurm,
        num_array_tasks: Optional[int],
        job_script_dir: str,
        job_name: str,
    ) -> str:
        num_array_tasks = None
        job_header = header_from_executor(
            job_name, executor, num_array_tasks, job_script_dir
        )
        return job_header

    def build(
        self, job: job_script.ClusterJob, job_name: str, job_script_dir: str
    ) -> str:
        assert isinstance(job.executor, executors.Slurm)
        assert isinstance(job.executable, executables.Command)
        return super().build(job, job_name, job_script_dir)


class SlurmHandle:
    def __init__(self, job_id: str) -> None:
        self.job_id = job_id

    async def wait(self):
        raise NotImplementedError()

    async def monitor(self):
        raise NotImplementedError()


def _slurm_job_predicate(job):
    if isinstance(job, xm.Job):
        return isinstance(job.executor, executors.Slurm)
    elif isinstance(job, array_job.ArrayJob):
        return isinstance(job.executor, executors.Slurm)
    else:
        raise ValueError(f"Unexpected job type: {type(job)}")


class SlurmClient(job_script.JobClient):
    builder_cls = SlurmJobScriptBuilder

    def __init__(
        self,
        settings: Optional[config_lib.ClusterSettings] = None,
        artifact_store: Optional[artifacts.ArtifactStore] = None,
    ) -> None:
        if settings is None:
            settings = config_lib.default().cluster_settings()
        self._settings = settings

        if artifact_store is None:
            project = config_lib.default().project()
            artifact_store = artifacts.create_artifact_store(
                settings.storage_root,
                hostname=settings.hostname,
                user=settings.user,
                project=project,
                connect_kwargs=settings.ssh_config,
            )
        self._artifact_store = artifact_store

        self._cluster = slurm.SlurmCluster(
            hostname=self._settings.hostname,
            username=self._settings.user,
        )

    def _launch(self, job_script_path, num_jobs):
        console.log(f"Launch with command:\n  sbatch {job_script_path}")
        cluster = self._cluster
        job_id = cluster.launch(job_script_path)
        console.log(f"Successfully launched job {job_id}")
        if num_jobs > 1:
            job_ids = [f"{job_id}_{i}" for i in range(num_jobs)]
        else:
            job_ids = [f"{job_id}"]
        return job_id, [SlurmHandle(j) for j in job_ids]


@functools.lru_cache()
def client() -> SlurmClient:
    return SlurmClient()


async def launch(job_name: str, job: job_script.ClusterJob) -> List[SlurmHandle]:
    if isinstance(job, array_job.ArrayJob):
        jobs = [job]  # type: ignore
    elif isinstance(job, xm.JobGroup):
        jobs: List[xm.Job] = xm.job_operators.flatten_jobs(job)
    elif isinstance(job, xm.Job):
        jobs = [job]

    jobs = [job for job in jobs if _slurm_job_predicate(job)]

    if not jobs:
        return []

    if len(jobs) > 1:
        raise ValueError(
            "Cannot launch a job group with multiple jobs as a single job."
        )

    if not isinstance(jobs[0].executor, executors.Slurm):
        raise ValueError(
            "Only GridEngine executors are supported by the gridengine backend."
        )

    return client().launch(job_name, jobs[0])


def _format_slurm_time(duration: datetime.timedelta) -> str:
    # See
    # https://github.com/SchedMD/slurm/blob/master/src/common/parse_time.c#L786
    days = duration.days
    seconds = int(duration.seconds)
    hours, remainder = divmod(seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    if days > 0:
        return "{:02}-{:02}:{:02}:{:02}".format(duration.days, hours, minutes, seconds)
    else:
        return "{:02}:{:02}:{:02}".format(hours, minutes, seconds)


def header_from_executor(
    job_name: str,
    executor: executors.Slurm,
    num_array_tasks: Optional[int],
    job_script_dir: str,
) -> str:
    header = []

    header.append(f"#SBATCH --job-name={job_name}")
    # TODO(yl): Only one task is supported for now.
    header.append("#SBATCH --ntasks=1")

    for resource, value in executor.resources.items():
        if value:
            header.append(f"#SBATCH --{resource}={value}")

    if executor.walltime is not None:
        duration = executor.walltime
        header.append(f"#SBATCH --time={_format_slurm_time(duration)}")

    log_directory = executor.log_directory or os.path.join(job_script_dir, "logs")
    if num_array_tasks is not None:
        stdout = os.path.join(log_directory, "slurm-%A_%a.out")
    else:
        stdout = os.path.join(log_directory, "slurm-%j.out")

    header.append(f"#SBATCH --output={stdout}")

    if executor.exclusive:
        header.append("#SBATCH --exclusive")

    if executor.partition:
        header.append(f"#SBATCH --partition={executor.partition}")

    if num_array_tasks is not None:
        array_spec = f"1-{num_array_tasks}"
        header.append(f"#SBATCH --array={array_spec}")

    # Skip requested header directives
    header = list(
        filter(
            lambda line: not any(skip in line for skip in executor.skip_directives),
            header,
        )
    )

    for line in executor.extra_directives:
        if not line.startswith("#SBATCH"):
            line = "#SBATCH " + line
        header.append(line)

    return "\n".join(header)



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common job operators useful in the framework and 3P-libraries."""

import copy
import itertools
from typing import Any, Callable, List, Sequence, Tuple

import attr
from lxm3._vendor.xmanager.xm import job_blocks
from lxm3._vendor.xmanager.xm import pattern_matching


def shallow_copy_job_type(
    job_type: job_blocks.JobTypeVar,
) -> job_blocks.JobTypeVar:
    """Creates a shallow copy of the job structure."""

    def apply_to_job_group(job_group: job_blocks.JobGroup) -> job_blocks.JobGroup:
        job_group = copy.copy(job_group)
        job_group.jobs = {key: matcher(job) for key, job in job_group.jobs.items()}
        return job_group

    matcher = pattern_matching.match(
        pattern_matching.Case([job_blocks.Job], copy.copy),
        apply_to_job_group,
        pattern_matching.Case(
            [job_blocks.JobGeneratorType], lambda generator: generator
        ),
        pattern_matching.Case([job_blocks.JobConfig], copy.copy),
    )
    return matcher(job_type)


def populate_job_names(job_type: job_blocks.JobTypeVar) -> None:
    """Assigns default names to the given jobs."""

    def apply_to_job(prefix: Sequence[str], target: job_blocks.Job) -> None:
        if target.name is None:
            target.name = "_".join(prefix) if prefix else target.executable.name

    def apply_to_job_group(prefix: Sequence[str], target: job_blocks.JobGroup) -> None:
        for key, job in target.jobs.items():
            matcher([*prefix, key], job)

    def ignore_unknown(_: Sequence[str], target: Any) -> None:
        return target

    matcher = pattern_matching.match(
        apply_to_job,
        apply_to_job_group,
        ignore_unknown,
    )
    return matcher([], job_type)


def collect_jobs_by_filter(
    job_group: job_blocks.JobGroup,
    predicate: Callable[[job_blocks.Job], bool],
) -> List[job_blocks.Job]:
    """Flattens a given job group and filters the result."""

    def match_job(job: job_blocks.Job) -> List[job_blocks.Job]:
        return [job] if predicate(job) else []

    def match_job_group(job_group: job_blocks.JobGroup) -> List[job_blocks.Job]:
        return list(
            itertools.chain.from_iterable(
                [job_collector(job) for job in job_group.jobs.values()]
            )
        )

    job_collector = pattern_matching.match(match_job_group, match_job)
    return job_collector(job_group)


@attr.s(auto_attribs=True)
class ConstraintClique:
    """A constraint with the list of jobs it applies to."""

    constraint: job_blocks.Constraint
    jobs: List[job_blocks.Job]


def aggregate_constraint_cliques(
    job_group: job_blocks.JobGroup,
) -> List[ConstraintClique]:
    """Forms constraint cliques.

    For each constraint met, collects all jobs it applies to.

    Args:
      job_group: A job group to aggregate on.

    Returns:
      A set of cliques.
    """

    def match_job(
        job: job_blocks.Job,
    ) -> Tuple[List[ConstraintClique], List[job_blocks.Job]]:
        return [], [job]

    def match_job_group(
        job_group: job_blocks.JobGroup,
    ) -> Tuple[List[ConstraintClique], List[job_blocks.Job]]:
        cliques: List[ConstraintClique] = []
        jobs: List[job_blocks.Job] = []
        for job in job_group.jobs.values():
            subcliques, subjobs = matcher(job)
            cliques += subcliques
            jobs += subjobs
        cliques = [
            ConstraintClique(constraint, jobs) for constraint in job_group.constraints
        ] + cliques
        return cliques, jobs

    matcher = pattern_matching.match(match_job, match_job_group)
    result, _ = matcher(job_group)
    return result


def flatten_jobs(job_group: job_blocks.JobGroup) -> List[job_blocks.Job]:
    return collect_jobs_by_filter(job_group, lambda _: True)



from unittest import mock

import fabric
from absl.testing import absltest
from absl.testing import parameterized

from lxm3.clusters import slurm


class SlurmTest(parameterized.TestCase):
    @parameterized.named_parameters(
        [
            {
                "testcase_name": "job",
                "text": "Submitted batch job 6",  # noqa
                "expected": 6,
            },
        ]
    )
    def test_parse_job_id(self, text, expected):
        job_id = slurm.parse_job_id(text)
        self.assertEqual(job_id, expected)

    def test_parse_invalid_job_id(self):
        with self.assertRaises(ValueError):
            slurm.parse_job_id("Failed")


class ClusterTest(absltest.TestCase):
    @mock.patch("fabric.Connection")
    def test_cluster(self, mock_connection):
        instance = mock_connection.return_value
        instance.run.return_value = fabric.Result(
            connection=instance,
            stdout="Submitted batch job 6",
        )
        cluster = slurm.SlurmCluster(hostname="host", username="user")
        job_id = cluster.launch("job.sbatch")
        self.assertEqual(job_id, 6)
        cluster.close()


if __name__ == "__main__":
    absltest.main()



from absl.testing import absltest
from absl.testing import parameterized

from lxm3 import xm
from lxm3 import xm_cluster
from lxm3._vendor.xmanager import xm_mock


class ArrayJobTest(parameterized.TestCase):
    @parameterized.parameters(
        (1,),
        (5,),
    )
    def test_array_job_empty_env_vars(self, num_tasks):
        args = [{"seed": 1}] * num_tasks
        job = xm_cluster.ArrayJob(
            xm_mock.MockExecutable(), xm_mock.MockExecutor(), args=args
        )
        self.assertEqual(job.args, list(map(xm.SequentialArgs.from_collection, args)))
        self.assertEqual(job.env_vars, [{}] * num_tasks)

    @parameterized.parameters(
        (1,),
        (5,),
    )
    def test_array_job_empty_args(self, num_tasks):
        env_vars = [{"TASK": "foo"}] * num_tasks
        job = xm_cluster.ArrayJob(
            xm_mock.MockExecutable(), xm_mock.MockExecutor(), env_vars=env_vars
        )
        self.assertEqual(job.args, [xm.SequentialArgs()] * num_tasks)
        self.assertEqual(job.env_vars, env_vars)

    def test_broadcast_args(self):
        args = [{"seed": 1}, {"seed": 2}]
        env_vars = [{"TASK": "foo"}]
        job = xm_cluster.ArrayJob(
            xm_mock.MockExecutable(),
            xm_mock.MockExecutor(),
            args=args,
            env_vars=env_vars,
        )
        self.assertEqual(
            list(map(lambda x: x.to_list(), job.args)), [["--seed=1"], ["--seed=2"]]
        )
        self.assertEqual(job.env_vars, env_vars * 2)

    def test_broadcast_env_vars(self):
        args = [{"seed": 1}]
        env_vars = [{"TASK": "foo"}, {"TASK": "bar"}]
        job = xm_cluster.ArrayJob(
            xm_mock.MockExecutable(),
            xm_mock.MockExecutor(),
            args=args,
            env_vars=env_vars,
        )
        self.assertEqual(
            list(map(lambda x: x.to_list(), job.args)), [["--seed=1"], ["--seed=1"]]
        )
        self.assertEqual(job.env_vars, env_vars)

    def test_invalid_broadcast_args(self):
        args = [{"seed": 1}, {"seed": 2}, {"seed": 3}]
        env_vars = [{"TASK": "foo"}, {"TASK": "bar"}]
        with self.assertRaises(ValueError):
            xm_cluster.ArrayJob(
                xm_mock.MockExecutable(),
                xm_mock.MockExecutor(),
                args=args,
                env_vars=env_vars,
            )


if __name__ == "__main__":
    absltest.main()



import itertools
import re
import shlex
import subprocess
import threading
import xml.etree.ElementTree as ET
from typing import Dict, List, Optional, Union

import fabric


def parse_job_id(output: str) -> re.Match:
    match = re.search(
        r"(?P<job_id>\d+)(.?(?P<task_start>\d+)-(?P<task_end>\d+):(?P<task_step>\d+))?",
        output,
    )
    if match is None:
        raise ValueError(f"Unable to parse job-id from:\n{output}")
    return match


def split_job_ids(match: re.Match) -> List[str]:
    job_id = match.group("job_id")
    task_start = match.group("task_start")
    task_end = match.group("task_end")
    task_step = match.group("task_step")
    if not task_start:
        return [job_id]
    else:
        return [
            f"{job_id}.{tid}"
            for tid in range(int(task_start), int(task_end) + 1, int(task_step))
        ]


def parse_qstat(qstat_xml_output: str):
    root = ET.fromstring(qstat_xml_output)
    infos = {}
    for joblist in root.iter("job_list"):
        job_number_elem = joblist.find("JB_job_number")
        if job_number_elem is None:
            raise ValueError()
        job_id = job_number_elem.text
        task_ids = joblist.find("tasks")
        if task_ids is not None:
            task_ids = task_ids.text
            if ":" in task_ids:  # type: ignore
                task_range, task_step = task_ids.split(":")  # type: ignore
                task_step = int(task_step)
                task_start, task_end = map(int, task_range.split("-"))
                for tid in range(task_start, task_end + 1, task_step):
                    infos[f"{job_id}.{tid}"] = {"state": joblist.attrib["state"]}
            elif "," in task_ids:  # type: ignore
                task_ids = map(int, task_ids.split(","))  # type: ignore
                for tid in task_ids:
                    infos[f"{job_id}.{tid}"] = {"state": joblist.attrib["state"]}
            else:
                tid = int(task_ids)  # type: ignore
                infos[f"{job_id}.{tid}"] = {"state": joblist.attrib["state"]}
        else:
            infos[job_id] = {"state": joblist.attrib["state"]}
    return infos


_KEYS = (
    "qname,hostname,group,owner,project,"
    "department,jobname,jobnumber,taskid,"
    "account,priority,qsub_time,start_time,"
    "end_time,granted_pe,slots,failed,exit_status,"
    "ru_wallclock,ru_utime,ru_stime,ru_maxrss,ru_ixrss,ru_ismrss,ru_idrss,ru_isrss,ru_minflt,"
    "ru_majflt,ru_nswap,ru_inblock,ru_oublock,ru_msgsnd,ru_msgrcv,ru_nsignals,ru_nvcsw,ru_nivcsw,"
    "cpu,mem,io,iow,maxvmem,arid,ar_sub_time,category"
).split(",")


def parse_accounting(data: str) -> List[Dict[str, str]]:
    records = []
    record = {}
    for i, line in enumerate(data.split("\n")):
        if line.startswith("="):
            if record:
                records.append(record)
            record = {}
        else:
            tokens = line.strip().split(maxsplit=1)
            if len(tokens) == 2:
                key, value = tokens
                if key in _KEYS:
                    record[key] = value
    if record:
        records.append(record)
    return records


def merge_taskids(task_ids: List[Union[int, str]]) -> List[str]:
    """Merge a list of task ids into list of ranges.

    This is useful for summarizing a list of (failed) task ids
    obtained via `qacct` to re-submit failed tasks.

    Examples:
    >>> merge_taskids([1, 2, 4, 5, 7])
    ['1-2', '4-5', '7']

    """
    ids = sorted([int(f) for f in task_ids])
    groups = itertools.groupby(
        ids, key=lambda item, c=itertools.count(): item - next(c)
    )
    tasks_str = []
    for range_ in [list(g) for k, g in groups]:
        if len(range_) == 1:
            tasks_str.append(str(range_[0]))
        else:
            tasks_str.append(f"{range_[0]}-{range_[-1]}")
    return tasks_str


class GridEngineCluster:
    _connection: Optional[fabric.Connection]

    def __init__(
        self, hostname: Optional[str] = None, username: Optional[str] = None
    ) -> None:
        self._hostname = hostname
        self._username = username
        self._connection = None
        self._mutex = threading.Lock()
        if hostname is not None:
            self._connection = fabric.Connection(host=hostname, user=username)
        self._qacct_conn = None

    def close(self):
        with self._mutex:
            if self._connection is not None:
                self._connection.close()
            if self._qacct_conn is not None:
                self._qacct_conn.close()

    def _run_command(self, command: str) -> str:
        if self._connection is None:
            return subprocess.check_output(shlex.split(command), text=True)
        else:
            with self._mutex:
                result = self._connection.run(command, hide="both")
                return result.stdout

    def launch(self, command):
        output = self._run_command(f"qsub {command}")
        match = parse_job_id(output)
        return match

    def qstat(self):
        stats = parse_qstat(self._run_command("qstat -xml"))
        return stats

    def cancel(self, job_id: str) -> None:
        self._run_command(f"qdel {job_id}")

    def qacct(self, job_id: Optional[str] = None):
        if self._qacct_conn is None:
            if self._connection is None:
                raise NotImplementedError("qacct with local session not implemented")
            sftp = self._connection.sftp()
            default_accounting_path = self._run_command(
                "echo $SGE_ROOT/$SGE_CELL/common/accounting"
            ).strip()
            try:
                sftp.stat(default_accounting_path)
            except IOError:
                act_qmaster_address = self._run_command(
                    "cat $SGE_ROOT/$SGE_CELL/common/act_qmaster"
                ).strip()
                qacct_conn = fabric.Connection(
                    act_qmaster_address, user=self._username, gateway=self._connection
                )
                self._qacct_conn = qacct_conn
            else:
                self._qacct_conn = self._connection

        with self._mutex:
            if not job_id:
                job_id = ""
            command = f"qacct -j {job_id} -u {self._username}"
            output = self._qacct_conn.run(command, hide="both").stdout
            return parse_accounting(output)

    def shell(self):
        if self._connection is None:
            raise NotImplementedError("shell with local session not implemented")
        self._connection.shell()

    def __repr__(self):
        if self._connection is not None:
            return f'Client(hostname="{self._hostname}", user="{self._username}")'
        else:
            return "Client()"



# https://github.com/sylabs/singularity/blob/main/internal/pkg/util/uri/uri_test.go
import appdirs
from absl.testing import absltest
from absl.testing import parameterized

from lxm3 import singularity


class URITest(parameterized.TestCase):
    # fmt: off
    @parameterized.named_parameters(
        ("docker basic", "docker://ubuntu", "sif", "ubuntu_latest.sif"),
        ("docker scoped", "docker://user/image", "oci.sif", "image_latest.oci.sif"),
        ("dave's magical lolcow", "docker://sylabs.io/lolcow", "sif", "lolcow_latest.sif"),
        ("docker w tags", "docker://sylabs.io/lolcow:3.7", "sif", "lolcow_3.7.sif"),
    )
    # fmt: on
    def test_get_name(self, uri, suffix, expected):
        n = singularity.uri.filename(uri, suffix)
        self.assertEqual(
            n,
            expected,
            msg='incorrectly parsed name as "%s" (expected "%s")' % (n, expected),
        )

    # fmt: off
    @parameterized.named_parameters(
        ("docker basic", "docker://ubuntu", "docker", "//ubuntu"),
        ("docker scoped", "docker://user/image", "docker", "//user/image"),
        ("dave's magical lolcow", "docker://sylabs.io/lolcow", "docker", "//sylabs.io/lolcow"),
        ("docker with tags", "docker://sylabs.io/lolcow:latest", "docker", "//sylabs.io/lolcow:latest"),
        ["library basic", "library://image", "library", "//image"],
        ("library scoped", "library://collection/image", "library", "//collection/image"),
        ("without transport", "ubuntu", "", "ubuntu"),
        ("without transport with colon", "ubuntu:18.04.img", "", "ubuntu:18.04.img"),
    )
    # fmt: on
    def test_split(self, uri, transport, ref):
        tr, r = singularity.uri.split(uri)
        self.assertEqual(
            (tr, r),
            (transport, ref),
            msg="incorrectly parsed uri as %s : %s (expected %s : %s)"
            % (tr, r, transport, ref),
        )

    def test_build_image_from_daemon(self):
        build_spec = "docker-daemon://jax-cuda:latest"
        transport, ref = singularity.uri.split(build_spec)
        self.assertEqual(transport, "docker-daemon")
        filename = singularity.uri.filename(build_spec, "sif")
        self.assertEqual(filename, "jax-cuda_latest.sif")
        appdirs.user_cache_dir("lxm3")


if __name__ == "__main__":
    absltest.main()



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import unittest

from lxm3._vendor.xmanager.xm import id_predictor
from lxm3._vendor.xmanager.xm import utils


class IdPredictorTest(unittest.TestCase):

  @utils.run_in_asyncio_loop
  async def test_first_id_is_correct(self):
    """Simple Predictor usage example."""
    predictor = id_predictor.Predictor(next_id=1)

    first_id = predictor.reserve_id()
    async with predictor.submit_id(first_id):
      self.assertEqual(first_id, 1)

  @utils.run_in_asyncio_loop
  async def test_ids_are_submitted_in_order(self):
    predictor = id_predictor.Predictor(next_id=1)

    self.assertEqual(predictor.reserve_id(), 1)
    self.assertEqual(predictor.reserve_id(), 2)
    self.assertEqual(predictor.reserve_id(), 3)

    submitted_ids = []

    async def submit(id_to_submit):
      async with predictor.submit_id(id_to_submit):
        submitted_ids.append(id_to_submit)

    await asyncio.gather(submit(3), submit(2), submit(1))

    self.assertEqual(submitted_ids, [1, 2, 3])

  @utils.run_in_asyncio_loop
  async def test_broken_sequence(self):
    predictor = id_predictor.Predictor(next_id=1)

    self.assertEqual(predictor.reserve_id(), 1)
    self.assertEqual(predictor.reserve_id(), 2)

    with self.assertRaises(RuntimeError):
      async with predictor.submit_id(1):
        raise RuntimeError('Id was eaten by a giant space ant.')

    with self.assertRaises(id_predictor.BrokenSequenceError):
      async with predictor.submit_id(2):
        pass


if __name__ == '__main__':
  unittest.main()



import os
import pathlib
import subprocess


def build_singularity_image(
    image_path: os.PathLike, build_spec: str, force: bool = True
):
    build_cmd = ["singularity", "build"]
    if force:
        build_cmd.append("--force")
    pathlib.Path(image_path).parent.mkdir(exist_ok=True, parents=True)
    cmd = build_cmd + [str(image_path), build_spec]
    subprocess.run(cmd, check=True)



import datetime
import glob
import os
import shutil
import subprocess
import tempfile

from lxm3.xm_cluster import executable_specs as cluster_executable_specs
from lxm3.xm_cluster.console import console

ENTRYPOINT_SCRIPT = "./entrypoint.sh"


class PackagingError(Exception):
    """Error raised when packaging fails."""


def _create_entrypoint_cmds(python_package: cluster_executable_specs.PythonPackage):
    if isinstance(python_package.entrypoint, cluster_executable_specs.ModuleName):
        cmds = ["python3 -m {}".format(python_package.entrypoint.module_name)]
    elif isinstance(python_package.entrypoint, cluster_executable_specs.CommandList):
        cmds = python_package.entrypoint.commands
    else:
        raise ValueError("Unexpected entrypoint: {}".format(python_package.entrypoint))
    cmds = "\n".join(cmds)
    # Allow passing extra parameters to the commands.
    if not cmds.endswith(("$@", '"$@"')):
        cmds = cmds + ' "$@"'
    return cmds


def create_python_archive(
    staging_directory: str, py_package: cluster_executable_specs.PythonPackage
):
    package_name = py_package.name
    version = datetime.datetime.now().strftime("%Y%m%d.%H%M%S")
    archive_name = f"{package_name}-{version}"
    package_dir = py_package.path
    resources = py_package.resources

    with tempfile.TemporaryDirectory() as tmpdir:
        with console.status(
            "Creating python package archive for {}".format(package_dir)
        ):
            try:
                subprocess.run(
                    [
                        "pip",
                        "install",
                        *py_package.pip_args,
                        "-t",
                        tmpdir,
                        package_dir,
                        *py_package.extra_packages,
                    ],
                    text=True,
                    check=True,
                    capture_output=True,
                )
            except subprocess.CalledProcessError as e:
                if e.stderr:
                    console.log("Error during packaging, stderr:", style="bold red")
                    console.log(e.stderr, style="bold red")
                raise PackagingError(
                    f"Failed to create python package from {package_dir}"
                ) from e

            # Add resources to the archive
            for resource in resources:
                for src, dst in resource.files:
                    target_file = os.path.join(tmpdir, dst)
                    if not os.path.exists(os.path.dirname(target_file)):
                        os.makedirs(os.path.dirname(target_file))
                    if not os.path.exists(target_file):
                        shutil.copy(src, target_file)
                    else:
                        raise ValueError(
                            "Additional resource overwrites existing file: %s", src
                        )

            if os.path.exists(os.path.join(tmpdir, "bin")):
                console.log('Removing "bin/" directory as these are not yet portable.')
                shutil.rmtree(os.path.join(tmpdir, "bin"))

            for f in glob.glob(os.path.join(tmpdir, "*.dist-info")):
                shutil.rmtree(f)

            entrypoint = f"""\
#!/bin/bash
export PYTHONPATH=$(dirname $0):$PYTHONPATH
{_create_entrypoint_cmds(py_package)}
"""
            with open(os.path.join(tmpdir, ENTRYPOINT_SCRIPT), mode="wt") as f:
                f.write(entrypoint)
            os.chmod(f.name, 0o755)
            archive_name = shutil.make_archive(
                os.path.join(staging_directory, archive_name),
                "zip",
                tmpdir,
                verbose=True,
            )

        console.log(
            f"Created archive: [repr.path]{os.path.basename(archive_name)}[repr.path]"
        )

    return os.path.basename(archive_name)


def create_universal_archive(
    staging_directory: str, universal_package: cluster_executable_specs.UniversalPackage
):
    version = datetime.datetime.now().strftime("%Y%m%d.%H%M%S")
    archive_name = f"{universal_package.name}-{version}"

    with tempfile.TemporaryDirectory(prefix=f"build-{archive_name}") as build_dir:
        package_dir = universal_package.path
        build_env = {**os.environ, "BUILDDIR": build_dir}

        build_script = os.path.join(package_dir, universal_package.build_script)

        try:
            subprocess.run(
                [build_script] + universal_package.build_args,
                cwd=package_dir,
                env=build_env,
                check=True,
            )
        except subprocess.CalledProcessError as e:
            raise PackagingError(
                "Failed to build universal package from {}".format(package_dir)
            ) from e

        archive_name = shutil.make_archive(
            os.path.join(staging_directory, archive_name),
            "zip",
            build_dir,
            verbose=True,
        )

    return os.path.basename(archive_name)



# type: ignore
import unittest.mock

from absl.testing import absltest
from absl.testing import parameterized

from lxm3.xm_cluster import config as config_lib

_SAMPLE_CONFIG = """
project = ""
[local]
[local.storage]
staging = ".lxm"

[[clusters]]
name = "cs"
server = "beaker.cs.ucl.ac.uk"
user = "foo"

[clusters.storage]
staging = "/home/foo/lxm3-staging"

[[clusters]]
name = "myriad"
server = "myriad.rc.ucl.ac.uk"
user = "ucaby36"

[clusters.storage]
staging = "/home/bar/Scratch/lxm3-staging"

"""


def _test_config():
    return config_lib.Config.from_string(_SAMPLE_CONFIG)


class ConfigTest(parameterized.TestCase):
    def test_config(self):
        config = _test_config()
        self.assertTrue(isinstance(config._data["clusters"], list))
        self.assertEqual(config._data["clusters"][0]["name"], "cs")

    def test_local_config(self):
        config = _test_config()
        self.assertEqual(config.local_settings().storage_root, ".lxm")

    def test_default_cluster(self):
        config = _test_config()
        self.assertEqual(config.default_cluster(), "cs")
        with unittest.mock.patch.dict("os.environ", {"LXM_CLUSTER": "myriad"}):
            self.assertEqual(config.default_cluster(), "myriad")

    def test_config_project(self):
        config = config_lib.Config.from_string("")
        self.assertEqual(config.project(), None)
        with unittest.mock.patch.dict("os.environ", {"LXM_PROJECT": "test"}):
            self.assertEqual(config.project(), "test")

    def test_local_settings(self):
        settings = config_lib.LocalSettings()
        self.assertEqual(settings.env, {})
        self.assertEqual(settings.singularity.env, {})
        self.assertEqual(settings.singularity.binds, {})

    def test_cluster_settings(self):
        settings = config_lib.ClusterSettings()
        self.assertEqual(settings.env, {})
        self.assertEqual(settings.singularity.env, {})
        self.assertEqual(settings.singularity.binds, {})


if __name__ == "__main__":
    absltest.main()
